import numpy as np


class Variable:
    def __init__(self, data, creator=None):
        if data is not None:
            if not isinstance(data, np.ndarray):
                raise TypeError(f"{type(data)} is not supported.")

        self.grad = None
        self.data = data
        self.creator = creator

    def backward(self):
        if self.grad is None:
            self.grad = np.ones_like(self.data)

        creators = [self.creator]
        while creators:
            creator_func = creators.pop()
            x, y = creator_func.input, creator_func.output
            x.grad = creator_func.backward(y.grad)

            if x.creator is not None:
                creators.append(x.creator)


class Function:
    def __call__(self, input_):
        y = self.forward(input_.data)
        output = Variable(as_array(y), creator=self)
        self.input = input_
        self.output = output
        return output

    def forward(self, x):
        raise NotImplementedError

    def backward(self, grad):
        raise NotImplementedError


def as_array(x):
    return np.asarray(x) if np.isscalar(x) else x



from deepzero.core import Exp, Square


class Exp(Function):
    def forward(self, x):
        return np.exp(x)

    def backward(self, grad):
        return grad * np.exp(self.input.data)


class Square(Function):
    def forward(self, x):
        return x ** 2

    def backward(self, grad):
        return grad * 2 * self.input.data


def square(x):
    return Square()(x)


def exp(x):
    return Exp()(x)
